﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;

namespace Watki
{
    class Program
    {
        static Random r = new Random();

        private static long[] liczbyTrafień;

        static void Main(string[] args)
        {
            int liczbaWatków = 5 * Environment.ProcessorCount;

            liczbyTrafień = new long[liczbaWatków];

            ParameterizedThreadStart ts = new ParameterizedThreadStart(uruchamianieObliczeńPi);
            Task[] tt = new Task[liczbaWatków];
            for (int i = 0; i < tt.Length; ++i)
            {
                Task t = new Task(uruchamianieObliczeńPi, i);
                tt[i] = t;
                t.Start();                
                Console.WriteLine("Obliczenia zostały uruchomione w zadaniu " + t.Id.ToString() + " (wątek główny: " + Thread.CurrentThread.ManagedThreadId + ")");
            }

            //czekanie na wątki
            Task.WaitAny(tt);
            Task.WaitAll(tt);            

            //obliczenia końcowe
            long całkowitaLiczbaTrafień = 0;
            for(int i = 0; i < tt.Length; ++i)
            {
                całkowitaLiczbaTrafień += liczbyTrafień[i];
                Console.WriteLine("Liczba trafień w wątku o indeksie " + i.ToString() + ": " + liczbyTrafień[i].ToString());
            }
            pi /= liczbaWatków; //liczona jakoś średnia średnich
            Console.WriteLine("Średnie pi {0}, błąd {1}", pi, Math.Abs(Math.PI - pi));
            pi = 4.0 * całkowitaLiczbaTrafień / 1000000000L;
            Console.WriteLine("Ostateczne pi {0}, błąd {1}", pi, Math.Abs(Math.PI - pi));
        }

        private static double pi = 0;

        static double obliczPi(long liczbaPrób, out long liczbaTrafień)
        {
            Random r = new Random(Program.r.Next() & DateTime.Now.Millisecond);
            double x, y;
            liczbaTrafień = 0;
            for (long i = 0; i < liczbaPrób; ++i)
            {
                /*
                //sztuczne
                if (i == liczbaPrób / 2)
                {
                    lock (Program.r)
                    {
                        Console.WriteLine("Symuluję np. zapis do pliku - START (" + Thread.CurrentThread.ManagedThreadId + ")");
                        Thread.Sleep(1000);
                        Console.WriteLine("Symuluję np. zapis do pliku - KONIEC (" + Thread.CurrentThread.ManagedThreadId + ")");
                    }
                }
                */

                x = r.NextDouble();
                y = r.NextDouble();
                if (x * x + y * y < 1) ++liczbaTrafień;
                //Console.WriteLine("x={0}, y={1}", x, y);
                if (i % (liczbaPrób / 100) == 0) Console.WriteLine(100.0 * i / liczbaPrób + "% (wątek: " + Thread.CurrentThread.ManagedThreadId + ", zadanie: " + Task.CurrentId + ")");
            }
            return 4.0 * liczbaTrafień / liczbaPrób;
        }

        static void uruchamianieObliczeńPi(object parametr)
        {
            try
            {
                int indeksWątku = (int)parametr;
                int startMs = Environment.TickCount;
                double pi = obliczPi(1000000000L / Environment.ProcessorCount / 5, out liczbyTrafień[indeksWątku]);
                Program.pi += pi;
                int endMs = Environment.TickCount;
                Console.WriteLine("Pi={0}, błąd={1}", pi, Math.Abs(Math.PI - pi));
                Console.WriteLine("Obliczenia trwały {0} ms", endMs - startMs);
                Console.WriteLine("Obliczenia przeprowadzone w wątku nr {0}", Thread.CurrentThread.ManagedThreadId);
            }
            catch (ThreadAbortException exc)
            {
                Console.WriteLine("Działanie wątku zostało przerwane (" + exc.Message + ") (wątek: " + Thread.CurrentThread.ManagedThreadId + ")");
            }
            catch (Exception exc)
            {
                Console.WriteLine("Wystąpił wyjątek (" + exc.Message + ")  (wątek główny: " + Thread.CurrentThread.ManagedThreadId + ")");
            }
        }
    }
}